/*
 * $Id: simpleproxy.c,v 1.27 1998/11/11 23:19:25 lord Exp $
 * ---------------------------------------------------------------------
 *
 * Simple proxy daemon 
 * ====================
 * Vadim Zaliva    <lord@crocodile.org>
 * Vlad  Karpinsky <vlad@crocodile.org>
 * Vadim Tymchenko <verylong@crocodile.org>
 * 
 * This software is provided ``as is'' without express or implied
 * warranty.  Feel free to distribute it.  Part of this code are taken
 * from W. RICHARD STEVENS "Unix Network Programming" excellent book.
 * 
 * This code should be considered alpha.  Feel free to contact us with
 * improvments requests and bug reports.
 * 
 * ---------------------------------------------------------------------
 *
 * $Log: simpleproxy.c,v $
 * Revision 1.27  1998/11/11 23:19:25  lord
 * Compiled under FreeBSD-2.2.7
 *
 * Revision 1.26  1998/10/27 23:49:34  vlad
 * I don't know is that a bug, but on Irix 6.2 parent process will not be able to accept any new connection if sockfd is closed in the child. Fixed.
 *
 * Revision 1.25  1998/09/15 00:59:34  vlad
 * version in the simpleproxy.c has been changed
 *
 * Revision 1.24  1998/09/14 04:49:48  lord
 * shutdown added.
 *
 * Revision 1.23  1998/09/10 22:33:54  vlad
 * Brush up. One minor bug fixed
 *
 * Revision 1.22  1998/08/24 20:48:56  lord
 * SO_REUSEADDR added
 *
 * Revision 1.21  1998/08/24 20:31:38  lord
 * exit on CTRLC added
 *
 * Revision 1.20  1998/08/24 20:27:53  lord
 * Compilation warning removed.
 * SIGINT handler added
 *
 * Revision 1.19  1998/04/01 19:40:23  verylong
 * Support of HTTPS proxy was added.
 * Crazy construction "if (isVerbose) ...." moved into function
 * log(format, ...)
 *
 * Revision 1.18  1997/12/30 06:55:28  lord
 * INETD works. RPM spec file added
 *
 * Revision 1.17  1997/12/29 23:05:20  lord
 * ident almoust done
 *
 * Revision 1.16  1997/12/26 20:26:33  lord
 * started inetd startup implementation
 *
 * Revision 1.15  1997/11/13 01:42:34  vlad
 * -D option
 *
 * Revision 1.14  1997/11/12 04:49:13  lord
 * POP3 proxy initial implementation
 *
 * Revision 1.13  1997/03/18 21:32:32  lord
 * Some preparation for final implementation
 * of new behaviour - according to CFG file.
 * Time to go back to California from S.Dacota.
 *
 * Revision 1.12  1997/03/18 20:39:46  lord
 * Now cfg file reads only if this options is specifyed.
 *
 * Revision 1.11  1997/03/18 18:01:03  lord
 * Spelling error in function name corrected.
 * Spelling error in man page corrected.
 * Added -7 flag for stripping 8th bit.
 *
 * Revision 1.10  1996/11/27 03:45:49  vlad
 * Writing PID to file added.
 *
 * Revision 1.9  1996/10/26 00:45:03  lord
 * Start implemnting cfg file.
 *
 * Revision 1.8  1996/10/17 23:07:47  vlad
 * Chield ending processing corrected.
 *
 * Revision 1.7  1996/10/16 18:39:32  lord
 * Now we unserstand symbolic port names from /etc/services.
 *
 * Revision 1.6  1996/10/16 18:25:42  vlad
 * Sintax error in version string fixed.
 *
 * Revision 1.5  1996/10/16 18:22:15  vlad
 * Check  for required options.
 *
 * Revision 1.4  1996/10/16 17:52:57  vlad
 * Output to stderr added.
 *
 * Revision 1.3  1996/10/16 05:59:10  lord
 * Parsin options with getopt done. Verbose and daemon flag.
 *
 * Revision 1.2  1996/10/16 02:06:22  vlad
 * Second fork() is removed. select() is used instead.
 *
 * Revision 1.1  1996/10/15 00:03:01  lord
 * Version 0
 *
 */

/* #define DEBUG 1 */

#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdarg.h>
#ifdef SUN
# include <sys/filio.h>
# include <stropts.h>
#endif
#include <sys/stat.h>
#include <syslog.h>
#include <netdb.h>
#include <fcntl.h>
#ifndef __FreeBSD__
# include <termio.h>
#endif
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#ifndef nil
# define nil NULL
#endif

#define MBUFSIZ         8192 

static char *SIMPLEPROXY_VERSION = "simpleproxy v2.2b by lord@crocodile.org,vlad@crocodile.org,verylong@crocodile.org";

struct cfg_record
{
    char *from;
    int  to_addr;
    int  to_port;
    struct cfg_record *next;
};

struct lst_record
{
    char *s;
    struct lst_record *next;
};

static void daemon_start(void);
static int  writen(int fd, char *ptr, int nbytes);
static void pass_all(int fd, int client);
static int  pass_one( int in, int out );
static int  gethostaddr(const char *name);
static int  readln(int fd, char *buf, int siz);
static void trim(char *str);
static struct lst_record * loadPOP3List(const char *popfile);
static int  checkPOP3List(struct lst_record *lst, char *acc);
static int  pop3login(int remotefd,int newsockfd,struct lst_record *poplist);
void        child_dead( int stat );
int         portN(const char *);
struct cfg_record *read_cfg(char *);
void        write_pid( char* filename );
int         open_remote(const char *PHost, int rportn,const char *remote_name);
static void logopen(void);
static void logclose(void);
static void log(char *format, ...);
static void ctrlc(int);
static int  HTTPSconnect(int remoteFd, char * remoteHost, int remotePort);

static int  isVerbose          = 0;
static int  isDaemon           = 0;
static int  isStripping        = 0;
static int  isStartedFromInetd = 0;
static long Delay              = 0;

static char * HTTPSproxy       = NULL;
static int  HTTPSport    = 0;

static int    sockfd=-1,newsockfd=-1,remotefd=-1;

int main(int ac, char **av)
{
    int    childpid;
    int    clien;
    struct sockaddr_in cli_addr,serv_addr;
    int    portn = -1 ,rportn = -1, rport;
    struct hostent *hp;
    char   *remote_name;
    extern char *optarg;
    extern int  optind;
    char  *PHost = nil;
    int   c;
    int   errflg=0;
    char  *cfgfile=nil;
    char  *popfile=nil;
    struct lst_record *poplist=nil;
    struct cfg_record *cfg=nil;
    char  *pidfile = nil;
    char *dest=nil;
    int rsp;
        
    
    while((c = getopt(ac, av, "iL:R:H:v7df:p:P:D:S:s:")) != -1)
	switch (c)
	{
	case 'v':
	    isVerbose++;
	    break;
	case 'i':
	    isStartedFromInetd++;
	    isDaemon++; 
	    break;
	case 'd':
	    isDaemon++;
	    break;
	case 'p':
	    pidfile = strdup(optarg);
	    break;
	case 'f':
	    cfgfile = strdup(optarg);
	    break;
	case 'L':
	    if(isStartedFromInetd)
		errflg++;
	    else
		portn = portN(optarg);
	    break;
	case 'P':
	    popfile = strdup(optarg);
	    break;
	case 'R':
	    rportn = portN(optarg);
	    break;
	case 'H':
	    PHost = strdup(optarg);
	    break;
	case 'D':
	    Delay = atol(optarg);
	    break;
	case '7':
	    isStripping=1;
	    break;
        case 'S':
            HTTPSproxy = strdup(optarg);
            break;
        case 's':
            HTTPSport =portN(optarg);
            break;
        default:
	    errflg++;
	}
    
    if(errflg || !PHost || rportn <= 0 || (portn<=0 && !isStartedFromInetd))
    {
	(void)fprintf(stderr,"%s\n",SIMPLEPROXY_VERSION);
	(void)fprintf(stderr, "usage:\n\tsimpleproxy [-d] [-p PIDfile] [-P <POP3 accounts list file>] [-f cfgfile] [-v] [-7] [-D delay] -L <local port> -H <remote host> -R <remote port> [-S <https_proxy> [-s <https proxy port>]]\n");
	exit(2);
    }


    logopen();
    log("Starting %s localhost:%d->%s:%d", SIMPLEPROXY_VERSION, portn, PHost, rportn);

    if(signal(SIGINT,ctrlc)==SIG_ERR)
        log("Error installing interrupt handler.");

    if(portn<=1024 && geteuid()!=0 && !isStartedFromInetd)
    {
        if (!isVerbose) {
            logopen();
            isVerbose++;
        }
        log("You must be root to run SIMPLEPROXY on reserved port");
        logclose();
	exit(1);
    }

    cfg=cfgfile?read_cfg(cfgfile):nil; 

    if(popfile)
	poplist=loadPOP3List(popfile);

    if(!isStartedFromInetd)
    {
	/* Let's become a daemon */
	if(isDaemon)
	    daemon_start(); 
	
	if(pidfile)
	    write_pid( pidfile );
	
	if((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
	{
            log("Error creating socket.");
            logclose();
	    exit(1);
	}
	
	memset((void *)&serv_addr,0,sizeof(serv_addr));
	serv_addr.sin_family=AF_INET;
	serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
	serv_addr.sin_port=htons(portn);

        if(setsockopt(sockfd,
		      SOL_SOCKET,
		      SO_REUSEADDR,
		      &rsp, sizeof(rsp)
        ))
            log("Error setting socket options");
        

	if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
	{
            log("Error binding socket.");
            logclose();
	    exit(1);
	}
        log("Waiting for connections.");

        if(listen(sockfd,5)<0)
        {
            log("Error listening socket: %s", strerror(errno));
            logclose();
            exit(1);
        }
	
	while(1)
	{
	    clien=sizeof(cli_addr);
	    newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clien);
	    
	    if(newsockfd<0)
	    {
		if ( errno ==  EINTR ) /* Interrupt after SIGCHLD */
		    continue;
                log("accept error: %s", strerror(errno));
                logclose();
		exit(1);
	    }
	    
	    signal( SIGCHLD, child_dead );
	    
	    childpid=fork();
	    if(childpid<0)
                log("error fork");
	    
	    if(childpid == 0) /* Child */
	    {
		hp=gethostbyaddr((char *)&cli_addr,clien,AF_INET);
		remote_name=strdup(hp?(hp->h_name):inet_ntoa(cli_addr.sin_addr));
                log("Got connection from: %s port %d",remote_name, portn);

/* I don't know is that a bug, but on Irix 6.2 parent process will not be abble
  accept any new connection if sockfd is closed here.      Vlad */
/*		(void)shutdown(sockfd,2); */
/*		(void)close(sockfd); */
		
		/* Process connection */
		
		if(cfg)
		{
		    /* Sreach in CFG */
		}
		
		if(!dest) { /* Not found or no CFG file */
                    dest=(HTTPSproxy) ? HTTPSproxy : PHost;
                    rport = (HTTPSproxy) ? HTTPSport : rportn;
                }
                remotefd=open_remote(dest, rport, remote_name);
                if(remotefd==-1)
                {
                    free(remote_name);
		    (void)shutdown(remotefd,2);
		    (void)shutdown(newsockfd,2);
                    (void)close(remotefd);
                    (void)close(newsockfd);
                    exit(1);
                }
                if (HTTPSproxy &&
                    HTTPSconnect(remotefd, PHost, rportn) != 0 ) {

                    free(remote_name);
                    (void)shutdown(remotefd,2);
                    (void)shutdown(newsockfd,2);
                    (void)close(remotefd);
                    (void)close(newsockfd);
                    exit(1);
                }
                free(PHost); /* We do not need it anymore */
                    
		if(poplist)
		{
		    /* Doing POP3 proxy */
		    if(pop3login(remotefd,newsockfd,poplist)==0)
			pass_all(remotefd,newsockfd); 
		} else {
		    /* Doing stream proxy */
		    pass_all(remotefd,newsockfd); 
		}
                log("Client %s done.",remote_name);

		free(remote_name);
                (void)shutdown(remotefd,2);
                (void)shutdown(newsockfd,2);
                (void)close(remotefd);
                (void)close(newsockfd);
		exit(0);
	    }
            close (newsockfd);
	}
    } else
    {
	/* Started from inetd */
	newsockfd=0;
	remote_name="INETD";

	if(cfg)
	{
	    /* Sreach in CFG */
	}
	
        if(!dest) { /* Not found or no CFG file */
            dest=(HTTPSproxy) ? HTTPSproxy : PHost;
            rport = (HTTPSproxy) ? HTTPSport : rportn;
        }

	remotefd=open_remote(dest, rportn, remote_name);
	if(remotefd==-1)
	{
            /* free(remote_name); */
            (void)shutdown(remotefd,2);
	    (void)close(remotefd);
	    exit(1);
	}
        if (HTTPSproxy &&
            HTTPSconnect(remotefd, PHost, rportn) != 0 ) {
            
            free(remote_name);
            (void)shutdown(remotefd,2);
            (void)shutdown(newsockfd,2);
            (void)close(remotefd);
            (void)close(newsockfd);
            exit(1);
        }
	free(PHost); /* We do not need it anymore */
	
	if(poplist)
	{
	    /* Doing POP3 proxy */
	    if(pop3login(remotefd,newsockfd,poplist)==0)
		pass_all(remotefd,newsockfd); 
	} else {
	    /* Doing stream proxy */
	    pass_all(remotefd,newsockfd); 
	}
        log("Client %s done.",remote_name);
        closelog();

	/* free(remote_name); */
        (void)shutdown(newsockfd,2);
	(void)close(newsockfd);
	exit(0);
    }
    
    /* NOTREACHED */
/*    if(isVerbose) */
    closelog();
    return(0);
}

/*
 * Write "n" bytes to a descriptor.
 * Use in place of write() when fd is a stream socket.
 */
static int writen(int fd, char *ptr, int nbytes)
{
    int	nleft, nwritten;

    nleft = nbytes;
    while (nleft > 0) {
	nwritten = write(fd, ptr, nleft);
	if (nwritten <= 0)
	    return(nwritten);		/* error */

	nleft -= nwritten;
	ptr   += nwritten;
    }
    return(nbytes - nleft);
}

/*
 * Detach a daemon process from login session context.
 */
static void daemon_start(void)
{

    /* Maybe I should do 2 forks here? */
    
    if (fork())
	exit(0);
    chdir("/");
    umask(0);
    (void) close(0);
    (void) close(1);
    (void) close(2);
    (void) open("/", O_RDONLY);
    (void) dup2(0, 1);
    (void) dup2(0, 2);
    setsid();
}


void pass_all( int fd, int client )
{
    fd_set         in;
    struct timeval tv;
    int            nsock, retval;
   
    if(fd > client)
	nsock    = fd + 1;
    else
	nsock    = client + 1;
    while(1)
    {
	tv.tv_sec   = 5;
	tv.tv_usec = 0;
	FD_ZERO( &in );
	FD_SET( fd, &in );
	FD_SET( client, &in );
	retval = select( nsock, &in, nil, nil, &tv );
	switch( retval )
	{
	case  0 :
	    /* Nothing to receive */
	    break;
	case -1:
	    /* Error occured */
	    return;
	default:
	    if ( FD_ISSET( fd, &in ))
		retval = pass_one( fd, client );
	    else if ( FD_ISSET( client, &in ))
		retval = pass_one( client, fd );
	    else
		retval = -1;
	    if ( retval < 0 )
		return;
	    if(Delay > 0)
		sleep(Delay);
	}
    }
}

static int gethostaddr(const char *name)
{
    static  int  res;
    struct hostent *he;
    int             a1,a2,a3,a4;
    
    if (sscanf(name,"%d.%d.%d.%d",&a1,&a2,&a3,&a4) == 4)
	return( inet_addr( name ) );
    else {
	he=gethostbyname(name);
	memcpy( &res , he->h_addr , he->h_length );
    }
    return( res );

}

static int  pass_one( int in, int out )
{
    int nread;
    char buff[MBUFSIZ];
    
    if((nread = readln(in, buff,MBUFSIZ)) <= 0)
	return -1;
    else
    {
        if(isStripping)
        {
            char *bufp;	
	    for(bufp=buff+nread-1;bufp>=buff;bufp--)
	        *bufp=*bufp&0177;
	}

	if (writen(out, buff, nread) != nread)
	{
            log("write error");
            return -1;
	}
    }
    return 0;
}

void child_dead( int stat )
{
    while ( waitpid( -1, nil, WNOHANG ) > 0 );
    signal( SIGCHLD, child_dead );
}

int  portN(const char *s)
{
    struct servent *se;
    
    if(s==nil)
	return -1; /* error */
    
    if(isdigit(*s))
	return atoi(s);
    
    else {
	if((se=getservbyname(s,"tcp"))==nil)
	    return -1;
	else
	    return  se->s_port;
    }
}

struct cfg_record *read_cfg(char *file)
{
    FILE *f;
    char str[2048];
    char *s;
    
    if((f=fopen(file,"r"))==nil)
    {
        log("Can't open cfg file: %s",file);
	return nil;
    }
    
    while((str==fgets(str,2040,f)))
    {
	s=str;
	if(*s=='#') continue; /* comment */
    }

    fclose(f);
    return nil;
}

void write_pid( char* filename )
{
    FILE *f;
    
    if((f=fopen(filename,"w"))==nil)
    {
        log("Can't open file '%s' to write PID",filename);
	return;
    }

    fprintf( f,"%d",getpid());
    fclose( f );
    return;
}

/**
  * Load list of allowed POP3 accounts from external file
  * One per line
  */
static struct lst_record *loadPOP3List(const char *popfile)
{
    FILE *f;
    char str[2048];
    char *s;
    struct lst_record *first = nil;
    struct lst_record *last  = nil;
    
    if((f=fopen(popfile,"r"))==nil)
    {
        log("Can't open POP3 file: %s",popfile);
	return nil;
    }

    while((str==fgets(str,2040,f)))
    {
	if(*str=='#') continue; /* comment */

	trim(str);
	if(*str=='\0') continue;
        log("Adding '%s' to POP3 users list",str);

	if(first==nil)
	{
	    first=malloc(sizeof(struct lst_record));
	    last=first;
	} else {
	    last->next=malloc(sizeof(struct lst_record));
	    last=last->next;
	}
	last->s=strdup(str);
	last->next=nil;
    }

    fclose(f);
    return first;
}

/**
  * Check if given account is OK to proxy
  */
static int checkPOP3List(struct lst_record *lst, char *acc)
{
    while(lst)
    {
	if(strcmp(lst->s,acc)==0)
	    return 1; /* found */
	else
	    lst=lst->next;
    }
    return 0;
}

static void trim(char *str)
{
    char *s;
    
    s=strchr(str,'\n'); if(s) *s='\0';
    s=strchr(str,' '); if(s) *s='\0';
    s=strchr(str,'\t'); if(s) *s='\0';
    s=strchr(str,'\r'); if(s) *s='\0';
}

static int  readln(int fd, char *buf, int siz)
{
    int  nread;
    
    nread = read(fd, buf, siz);
    if (nread <= 0)
    {
	if (nread < 0)
	    log("read error");
	return -1;
    } else
    {
	return nread;
    }
}

/*
 * Pass USER command to remote end
 * only if user is in the list
 *
 * TODO:
 * There is potential security problem in this implementation
 * User might send another USER comand after first failed authentification.
 * Need to be addressed.
 *
 * See RFC1725 for details.
 */
static int pop3login(int remotefd,int newsockfd,struct lst_record *poplist)
{
    static char errmsg[]={"-ERR Not allowed by proxy\r\n"};
    static char errmsg1[]={"-ERR Can't get your user name\r\n"};
    char buff[MBUFSIZ];
    char *s;
    int  n;
    
    /* read +OK from server */

    if(readln(remotefd,buff,MBUFSIZ)<=0) return 1; /* read error */
    if(strncmp(buff,"+OK",3)        !=0) return 1; /* server is not OK */
    
    /* Send client our OK */
    writen(newsockfd, "+OK ",4);
    writen(newsockfd, SIMPLEPROXY_VERSION, strlen(SIMPLEPROXY_VERSION));
    writen(newsockfd, "\r\n",2);

    /* read cmd from client */

    if((n=readln(newsockfd,buff,MBUFSIZ))<=0) return 1; /* read error */
    do {
	buff[n]='\0';
	n--;
    } while((buff[n]=='\r' || buff[n]=='\n') && n>=0);
    if(n==0) return 1; /* empty CMD */
						  
    if(strncmp(buff,"USER ",5)       !=0) return 1; /* first command is not USER */

    /* get user name */
    s=strdup(buff+5);
    trim(s);
    if(*s=='\0')
    {
	free(s);
	writen(newsockfd,errmsg1,strlen(errmsg1)); /* Send error to client */
	return 1; /* invalid user name */
    }
    
    /* Search user in access list */
    if(checkPOP3List(poplist,s)==0)
    {
	free(s);
	writen(newsockfd,errmsg,strlen(errmsg)); /* Send error to client */
	return 1; /* user not found */
    }

    /* forward USER command to server */
    writen(remotefd,buff,strlen(buff));
    writen(remotefd,"\r\n",2);
    
    return 0; /* ok */
}
 
int open_remote(const char *PHost, int rportn, const char *remote_name)
{
    struct sockaddr_in remote_addr;
    int remotefd;
    
    if((remotefd=socket(AF_INET,SOCK_STREAM,0))==-1)
    {
        log("Can't create socket. Terminating client %s done.",remote_name);
	return -1;
    }
    
    remote_addr.sin_family=AF_INET;
    remote_addr.sin_addr.s_addr=gethostaddr(PHost);
    remote_addr.sin_port=htons(rportn);
    
    if(connect(remotefd, (struct sockaddr *) &remote_addr, sizeof(remote_addr)))
	return -1;
    
    log("Made connection to: %s port %d",PHost, rportn );
    return remotefd;
}

static int HTTPSconnect(int remoteFd, char * remoteHost, int remotePort)
{
    char buff[MBUFSIZ];
    char *s;
    long  n;

    /* prepare command and connect to remote side */
    sprintf(buff, "CONNECT %s:%i HTTP/1.0\nUser-agent: %s\n\n",
            remoteHost, remotePort, SIMPLEPROXY_VERSION);
    n = strlen(buff);
    if (writen(remoteFd, buff, n) != n) return 1; /* write error */

    /* reading responce from the server */
    if(readln(remoteFd,buff,MBUFSIZ)<=0) return 1; /* read error */

    /* Format of HTTPS proxy responce must be: <Protocol> <code> <message> */
    if (!(s = strchr(buff, ' '))) return 1;
    if ((n = strtol(s, &s, 10)) != 200) {
        log("Can't establish connection through HTTPS proxy. Proxy responds:\n %s", buff);
        return 1;
    }
    return 0; /* ok */
}

static void logopen(void)
{
    if (isVerbose & isDaemon)
        openlog("simpleproxy", LOG_PID| LOG_CONS|LOG_NOWAIT, LOG_USER);

}

static void logclose(void)
{
    if (isVerbose && isDaemon)
        closelog();
}

static void log(char *format, ...)
{
    if(isVerbose) {
        va_list ap;
        va_start(ap, format);

        if(isDaemon) {
            char buffer[MBUFSIZ];
            
            (void) vsprintf(buffer, format, ap);
	    syslog(LOG_INFO,buffer);
	} else {
            (void) fprintf(stderr, "simpleproxy[%d]:", (int)getpid());
            (void) vfprintf(stderr, format, ap);
            (void) fprintf(stderr, "\n");
        }
        va_end(ap);
    }
}

static void ctrlc(int s)
{
    log("Interrupted... Shutting down connections");
    
    if(sockfd    !=-1) 
    {
	(void)shutdown(sockfd,2);
	(void)close(sockfd   );
    }
    if(newsockfd !=-1)
    {
	(void)shutdown(newsockfd,2);
	close(newsockfd);
    }
    if(remotefd  !=-1)
    {
	(void)shutdown(remotefd,2);
	close(remotefd );
    }

    /* system V style. */
    if(signal(SIGINT,ctrlc)==SIG_ERR)
        log("Error installing interrupt handler.");
    exit(1);
}
